/*
** Snake pe LCD
** George Matei 332CA
** PM 2016
** 
*/
#define F_CPU 12000000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BOARD_START_X 2
#define BOARD_START_Y 11

#define CELL_SIZE 4

#define BOARD_WIDTH	22
#define BOARD_HEIGHT 14

/* Array cu starea board-ului de joc */
static unsigned char game_board[BOARD_WIDTH][BOARD_HEIGHT];

#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3
#define NONE 4
#define FOOD 5

typedef struct coord {
	unsigned char x, y;
} coord;

static coord head, tail;
static int dir = 3;

/* Coordonate relative pentru deplasare */
static char dx[4] = {0, 0, -1, +1};
static char dy[4] = {-1, +1, 0, 0};

static int score = 0;


/* Deseneaza in screenbuffer o celula din game_board */
void draw_board_cell(unsigned char i, unsigned char j)
{
	char x, y, k, l;

	x = BOARD_START_X + i * CELL_SIZE;
	y = BOARD_START_Y + j * CELL_SIZE;

	/* Desenez ceilalti pixelii */
	switch (game_board[i][j]) {

		case UP: case DOWN:
			for (k = y; k < y + CELL_SIZE; k++) {
				lcd_set_pixel(x + 1, k);
				lcd_set_pixel(x + 2, k);
			}
			break;

		case LEFT : case RIGHT:
			for (k = x; k < x + CELL_SIZE; k++) {
				lcd_set_pixel(k, y + 1);
				lcd_set_pixel(k, y + 2);
			}
			break;

		case NONE:
			for (k = x; k < x + CELL_SIZE; k++)
				for (l = y; l < y + CELL_SIZE; l++)
					lcd_clear_pixel(k, l);
			break;

		case FOOD:
			for (k = x + 1; k < x + 3; k++) {
				for (l = y + 1; l < y + 3; l++) {
					lcd_set_pixel(k, l);
				}
			}
			break;
	}
}

/* Deseneaza o celula in care s-a luat o curba */
void draw_turning_cell(unsigned char i, unsigned char j, unsigned char tip)
{
	char x, y, k, l;

	x = BOARD_START_X + i * CELL_SIZE;
	y = BOARD_START_Y + j * CELL_SIZE;

	/* Mai intai sterg celula */
	for (k = x; k < x + CELL_SIZE; k++)
		for (l = y; l < y + CELL_SIZE; l++)
			lcd_clear_pixel(k, l);

	/* Setez centrul */
	for (k = 1; k < 3; k++)
		for (l = 1; l < 3; l++)
			lcd_set_pixel(x + k, y + l);

	switch (tip) {
		case UP:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x, y + k);
				lcd_set_pixel(x + k, y + 3);
			}
			break;

		case DOWN:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x + 3, y + k);
				lcd_set_pixel(x + k, y + 3);
			}
			break;

		case LEFT:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x, y + k);
				lcd_set_pixel(x + k, y);
			}
			break;

		case RIGHT:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x + k, y);
				lcd_set_pixel(x + 3, y + k);
			}
			break;
	}

}


void draw_game_board()
{
	for (unsigned char i = 0; i < BOARD_WIDTH; i++) {
		for (unsigned char j = 0; j < BOARD_HEIGHT; j++) {
			draw_board_cell(i, j);
		}
	}
}

/* Plaseaza un FOOD aleator */
void generate_food()
{
	unsigned char x, y;
	do {
		x = rand() % BOARD_WIDTH + BOARD_START_X;
		y = rand() % BOARD_HEIGHT + BOARD_START_Y;
	} while(game_board[x][y] != NONE);
	game_board[x][y] = FOOD;
	draw_board_cell(x, y);
}

/* Initializeaza jocul */
void init_game_board()
{
	unsigned char i, j;
	for (i = 0; i < BOARD_WIDTH; i++)
		for (j = 0; j < BOARD_HEIGHT; j++)
			game_board[i][j] = NONE;

	// Initializare head si tail
	head.x = BOARD_WIDTH / 2 + 2;
	tail.x = BOARD_WIDTH / 2 - 2;
	head.y = tail.y = BOARD_HEIGHT / 3;

	// Desenare sarpe
	for (i = tail.x; i <= head.x; i++) {
		game_board[i][head.y] = RIGHT;
	}

	generate_food();

}

void game_over()
{	
	lcd_clear();
	lcd_goto_xy(1, 3);
	printf("Game over");
	lcd_goto_xy(1, 4);
	printf("Press reset to retry!");
	while (1);
}

/* Updateaza valoarea fiecarui pixel in functie de directia de mers */
unsigned char update_value(unsigned char v, char dv, unsigned char max)
{
	if (v + dv == max) {
		v = 0;
		return v;
	} else if (v == 0 && dv == -1) {
		v = max - 1;
		return v;
	}

	v += dv;
	return v;
}

/* Muta sarpele inainte */
void go_forward()
{
	unsigned char current_tail, current_head;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];
	
	game_board[tail.x][tail.y] = NONE;

	/* Avansez capul */
	head.x = update_value(head.x, dx[current_head], BOARD_WIDTH);
	head.y = update_value(head.y, dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 10;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			tail.x = update_value(tail.x, dx[current_tail], BOARD_WIDTH);
			tail.y = update_value(tail.y, dy[current_tail], BOARD_HEIGHT);
		} else {
			game_over();
		}
	}

	game_board[head.x][head.y] = current_head;

	draw_board_cell(head.x, head.y);
}

/* Turn helper */
static unsigned char directii[2][4] = {{0, 3, 1, 2}, {1, 2, 3, 0}};
/* Coordonate relative pentru deplasarea la stanga */
static char left_dx[4] = {-1, 1, 0, 0};
static char left_dy[4] = {0, 0, 1, -1};
static char left_next[4] = {LEFT, RIGHT, DOWN, UP};

/* Muta sarpele la stanga */
void go_left()
{
	unsigned char current_head, current_tail;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];

	game_board[tail.x][tail.y] = NONE;
	game_board[head.x][head.y] = left_next[current_head];

	/* Avansez capul */
	draw_turning_cell(head.x, head.y, directii[0][current_head]);
	head.x = update_value(head.x, left_dx[current_head], BOARD_WIDTH);
	head.y = update_value(head.y, left_dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 10;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			tail.x = update_value(tail.x, dx[current_tail], BOARD_WIDTH);
			tail.y = update_value(tail.y, dy[current_tail], BOARD_HEIGHT);
		} else { // Este o bucata din sarpe; GAME OVER.
			game_over();
		}
	}

	game_board[head.x][head.y] = left_next[current_head];

	draw_board_cell(head.x, head.y);
}

/* Coordonate relative pentru deplasare la dreapta */
static char right_dx[4] = {1, -1, 0, 0};
static char right_dy[4] = {0, 0, -1, +1};
static char right_next[4] = {RIGHT, LEFT, UP, DOWN};

/* Muta sarpele la stanga */
void go_right()
{
	unsigned char current_head, current_tail;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];

	game_board[tail.x][tail.y] = NONE;
	game_board[head.x][head.y] = right_next[current_head];

	/* Avansez capul */
	draw_turning_cell(head.x, head.y, directii[1][current_head]);
	head.x = update_value(head.x, right_dx[current_head], BOARD_WIDTH);
	head.y = update_value(head.y, right_dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 10;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			tail.x = update_value(tail.x, dx[current_tail], BOARD_WIDTH);
			tail.y = update_value(tail.y, dy[current_tail], BOARD_HEIGHT);
		} else { // Game over
			game_over();
		}
	}

	game_board[head.x][head.y] = right_next[current_head];

	draw_board_cell(head.x, head.y);
}

/* Setez cele 4 butoane ca input */
void input_init()
{
	DDRD &= ~ (1 << PD1);
	PORTD |= (1 << PD1);
	
	DDRD &= ~ (1 << PD4);
	PORTD |= (1 << PD4);
	
	DDRD &= ~ (1 << PD0);
	PORTD |= (1 << PD0);
	
	DDRD &= ~ (1 << PD5);
	PORTD |= (1 << PD5);
	
}

void draw_score()
{
	lcd_goto_xy(1,1);
	char buff[20];
	itoa(score,buff,10);
	lcd_str(buff);
}

static char pressed_left = 0;
static char pressed_right = 0;
static char pressed_up = 0;
static char pressed_down = 0;

int main(void) {
	
	input_init();

	lcd_init();
	lcd_contrast(0x40);

	lcd_clear();
	srand(3);
	
	while (1) {
		lcd_goto_xy(1,1);
		lcd_str("Press any key to start the  ");
		lcd_str("game");
		// UP
        if ((PIND & (1 << PD0)) == 0) {
			break;
		}
    
		// LEFT
		if ((PIND & (1 << PD1)) == 0) {
			break;
		}
    
		// RIGHT
		if ((PIND & (1 << PD4)) == 0) {
			break;
			
		}
    
		// DOWN
		if ((PIND & (1 << PD5)) == 0) {
			break;
		}
		_delay_ms(20);
		lcd_clear();
    }
	
	init_game_board();
	draw_game_board();
	
	while (1) {
		lcd_clear();
		if (dir == RIGHT)
			if (pressed_up == 1) {
				go_left();
				dir = UP;
				pressed_up = 0;
			} else {
				if (pressed_down == 1) {
					go_right();
					dir = DOWN;
					pressed_down = 0;
					
				} else {
					go_forward();
				}
			}
		else 
			if (dir == LEFT) {
				if (pressed_up == 1) {
					go_right();
					pressed_up = 0;
					dir = UP;
				} else {
					if (pressed_down == 1) {
						go_left();
						pressed_down = 0;
						dir = DOWN;
					} else {
						go_forward();
					}
				}
			}
			else 
				if (dir == UP) {
					if(pressed_right == 1) {
						go_right();
						pressed_right = 0;
						dir = RIGHT;
					} else {
						if (pressed_left == 1) {
							go_left();
							pressed_left = 0;
							dir = LEFT;
						} else {
							go_forward();
						}
					}
					
				}
				else
					if (dir == DOWN) {
						if(pressed_right == 1) {
							go_left();
							pressed_right = 0;
							dir = RIGHT;
						} else {
							if (pressed_left == 1) {
								go_right();
								pressed_left = 0;
								dir = LEFT;
							} else {
								go_forward();
							}
						}
					}
					
		lcd_draw_border();
		lcd_draw_buffer();
		draw_game_board();
		draw_score();
		
		pressed_right = pressed_left = pressed_up = pressed_down = 0;
		for (int i = 0; i < 10000; i++) {
			for (int j = 0; j < 30; j++) {
				// right
				if ((PIND & (1 << PD4)) == 0) {
					pressed_right = 1;
				}
				
				// left 
				if ((PIND & (1 << PD1)) == 0) {
					pressed_left = 1;
				}
				
				// up
				if ((PIND & (1 << PD0)) == 0) {
					pressed_up = 1;
				}
				
				// down
				if ((PIND & (1 << PD5)) == 0) {
					pressed_down = 1;
				}
			}
		}
	}
}
